__author__ = 'Joseph Ryan'
__license__ = "GPLv2"
__maintainer__ = "Joseph Ryan"
__email__ = "jr@aphyt.com"

from aphyt.eip import *
from decimal import *


def two_bytes_to_fixed_point_temperature(two_bytes: bytes):
    temperature = struct.unpack("<H", two_bytes)[0]
    temperature = Decimal(temperature) / Decimal(10.0)
    return temperature


def two_bytes_to_uint(two_bytes: bytes):
    uint_instance = struct.unpack("<H", two_bytes)[0]
    uint_instance = Decimal(uint_instance) / Decimal(10.0)
    return uint_instance


def unsigned_integer_to_temp(unsigned: int):
    temperature = Decimal(unsigned) / Decimal(10.0)
    return temperature


class SensorMonitorObject:
    def __init__(self):
        self.sensor_version = None
        self.sensor_status = None
        self.alarm_status = None
        self.internal_temperature_value = None
        self.internal_max_temperature_value = None
        self.internal_predicted_arrival_time = None
        self.segment_temperature_list = [None] * 16
        self.segment_max_temperature_list = [None] * 16
        self.segment_predicted_temperature = [None] * 16
        self._temp_list_offset = 12
        self._max_temp_list_offset = self._temp_list_offset + 16
        self._predicted_temp_list_offset = self._max_temp_list_offset

    def from_bytes(self, sensor_monitor_attributes_bytes: bytes):
        self.sensor_version = two_bytes_to_uint(sensor_monitor_attributes_bytes[0:2])
        self.sensor_status = two_bytes_to_uint(sensor_monitor_attributes_bytes[2:4])
        self.alarm_status = two_bytes_to_uint(sensor_monitor_attributes_bytes[4:6])
        self.internal_temperature_value = two_bytes_to_fixed_point_temperature(sensor_monitor_attributes_bytes[6:8])
        self.internal_max_temperature_value = \
            two_bytes_to_fixed_point_temperature(sensor_monitor_attributes_bytes[8:10])
        self.internal_predicted_arrival_time = \
            two_bytes_to_fixed_point_temperature(sensor_monitor_attributes_bytes[10:12])
        for i in range(16):
            t_start = i*2 + self._temp_list_offset
            mt_start = i*2 + self._max_temp_list_offset
            pt_start = i*2 + self._predicted_temp_list_offset
            temp_bytes = \
                sensor_monitor_attributes_bytes[t_start:t_start+2]
            self.segment_temperature_list[i] = two_bytes_to_fixed_point_temperature(temp_bytes)
            max_bytes = \
                sensor_monitor_attributes_bytes[mt_start:mt_start+2]
            self.segment_max_temperature_list[i] = two_bytes_to_fixed_point_temperature(max_bytes)
            predict_bytes = \
                sensor_monitor_attributes_bytes[pt_start:pt_start+2]
            self.segment_predicted_temperature[i] = two_bytes_to_fixed_point_temperature(predict_bytes)


class K6PMTH:
    """
    CIP Class ID


    """
    def __init__(self):
        super().__init__()
        self.connected_cip_dispatcher = EIPConnectedCIPDispatcher()

    def connect_explicit(self, host):
        # ToDo Consider moving all these convenience methods to a Mixin
        self.connected_cip_dispatcher.connect_explicit(host)

    def close_explicit(self):
        self.connected_cip_dispatcher.close_explicit()

    def register_session(self):
        self.connected_cip_dispatcher.register_session()

    def get_attribute_single_as_uint(self, read_address: address_request_path_segment):
        cip_unsigned_integer = CIPUnsignedInteger()
        cip_unsigned_integer.data = self.connected_cip_dispatcher.get_attribute_single_service(read_address).reply_data
        return cip_unsigned_integer.value()

    def get_attribute_single_as_temp(self, read_address: address_request_path_segment):
        cip_unsigned_integer = CIPUnsignedInteger()
        cip_unsigned_integer.data = self.connected_cip_dispatcher.get_attribute_single_service(read_address).reply_data
        temperature = Decimal(cip_unsigned_integer.value()) / Decimal(10.0)
        return temperature

    def main_unit_status(self):
        read_address = address_request_path_segment(class_id=b'\x74\x03', instance_id=b'\x01', attribute_id=b'\x64')
        main_unit_status = CIPUnsignedInteger()
        main_unit_status.data = self.connected_cip_dispatcher.get_attribute_single_service(read_address).reply_data
        return main_unit_status

    def running_time(self):
        read_address = address_request_path_segment(class_id=b'\x74\x03', instance_id=b'\x01', attribute_id=b'\x65')
        running_time = CIPUnsignedInteger()
        running_time.data = self.connected_cip_dispatcher.get_attribute_single_service(read_address).reply_data
        return running_time

    def software_version(self):
        read_address = address_request_path_segment(class_id=b'\x74\x03', instance_id=b'\x01', attribute_id=b'\x66')
        software_version = CIPUnsignedInteger()
        software_version.data = self.connected_cip_dispatcher.get_attribute_single_service(read_address).reply_data
        return software_version

    def number_of_connected_sensors(self):
        read_address = address_request_path_segment(class_id=b'\x74\x03', instance_id=b'\x01', attribute_id=b'\x67')
        sensor_count = CIPUnsignedInteger()
        sensor_count.data = self.connected_cip_dispatcher.get_attribute_single_service(read_address).reply_data
        return sensor_count
    
    def sensor_in_position_adjustment_mode(self):
        read_address = address_request_path_segment(class_id=b'\x74\x03', instance_id=b'\x01', attribute_id=b'\x68')
        sensor_in_position_adjustment_mode = CIPUnsignedInteger()
        sensor_in_position_adjustment_mode.data = self.connected_cip_dispatcher.get_attribute_single_service(read_address).reply_data
        return sensor_in_position_adjustment_mode

    def sensor_monitor_object(self, sensor_number: int):
        instance_id = struct.pack("<B", sensor_number)
        read_address = address_request_path_segment(class_id=b'\x75\x03', instance_id=instance_id)
        sensor_monitor_object = SensorMonitorObject()
        sensor_monitor_object.from_bytes(self.connected_cip_dispatcher.get_attribute_all_service(read_address).reply_data)
        return sensor_monitor_object

    def sensor_version(self, sensor_number: int):
        instance_id = struct.pack("<B", sensor_number)
        read_address = address_request_path_segment(class_id=b'\x75\x03', instance_id=instance_id, attribute_id=b'\x64')
        sensor_version = CIPUnsignedInteger()
        sensor_version.data = self.connected_cip_dispatcher.get_attribute_single_service(read_address).reply_data
        return sensor_version

    def sensor_status(self, sensor_number: int):
        instance_id = struct.pack("<B", sensor_number)
        read_address = address_request_path_segment(class_id=b'\x75\x03', instance_id=instance_id, attribute_id=b'\x65')
        sensor_status = CIPUnsignedInteger()
        sensor_status.data = self.connected_cip_dispatcher.get_attribute_single_service(read_address).reply_data
        return sensor_status

    def sensor_alarm_status(self, sensor_number: int):
        instance_id = struct.pack("<B", sensor_number)
        read_address = address_request_path_segment(class_id=b'\x75\x03', instance_id=instance_id, attribute_id=b'\x66')
        sensor_alarm_status = CIPUnsignedInteger()
        sensor_alarm_status.data = self.connected_cip_dispatcher.get_attribute_single_service(read_address).reply_data
        return sensor_alarm_status

    def internal_temperature(self, sensor_number: int):
        instance_id = struct.pack("<B", sensor_number)
        read_address = address_request_path_segment(class_id=b'\x75\x03', instance_id=instance_id, attribute_id=b'\x67')
        return self.get_attribute_single_as_temp(read_address)

    def internal_maximum_temperature(self, sensor_number: int):
        instance_id = struct.pack("<B", sensor_number)
        read_address = address_request_path_segment(class_id=b'\x75\x03', instance_id=instance_id, attribute_id=b'\x68')
        return self.get_attribute_single_as_temp(read_address)

    def internal_predicted_arrival(self, sensor_number: int):
        instance_id = struct.pack("<B", sensor_number)
        read_address = address_request_path_segment(class_id=b'\x75\x03', instance_id=instance_id, attribute_id=b'\x69')
        return self.get_attribute_single_as_temp(read_address)

    def segment_temp_current(self, sensor_number: int, segment_number: int):
        instance_id = struct.pack("<B", sensor_number)
        attribute_id = struct.pack("<B", segment_number + 0x6a)
        read_address = address_request_path_segment(
            class_id=b'\x75\x03', instance_id=instance_id, attribute_id=attribute_id)
        return self.get_attribute_single_as_temp(read_address)

    def segment_temp_max(self, sensor_number: int, segment_number: int):
        instance_id = struct.pack("<B", sensor_number)
        attribute_id = struct.pack("<B", segment_number + 0x7a)
        read_address = address_request_path_segment(
            class_id=b'\x75\x03', instance_id=instance_id, attribute_id=attribute_id)
        return self.get_attribute_single_as_temp(read_address)

    def segment_temp_predicted(self, sensor_number: int, segment_number: int):
        instance_id = struct.pack("<B", sensor_number)
        attribute_id = struct.pack("<B", segment_number + 0x8a)
        read_address = address_request_path_segment(
            class_id=b'\x75\x03', instance_id=instance_id, attribute_id=attribute_id)
        return self.get_attribute_single_as_temp(read_address)

    def pixel_temperatures(self, sensor_number: int):
        instance_id = struct.pack("<B", sensor_number)
        temporary_array = CIPArray()
        temp_instance = CIPUnsignedInteger()
        temporary_array.from_instance(temp_instance, temp_instance.size, 1, [64], [0])
        result_array = []
        for i in range(16):
            attribute_id = struct.pack("<B", i + 0x64)
            read_address = address_request_path_segment(
                class_id=b'\x76\x03', instance_id=instance_id, attribute_id=attribute_id)
            temporary_array.data = self.connected_cip_dispatcher.get_attribute_single_service(read_address).reply_data
            result_array.append(temporary_array.value())
        return result_array



__author__ = 'Joseph Ryan'
__license__ = "GPLv2"
__maintainer__ = "Joseph Ryan"
__email__ = "jr@aphyt.com"

from ..eip import *


class TCPInterfaceObject(CIPObject, GetAttributeSingleMixin, SetAttributeSingleMixin, GetAttributeAllMixin):
    def __init__(self, cip_dispatcher: EIPConnectedCIPDispatcher):
        super().__init__(b'\xf5')
        self.instance_id = b'\x01'
        self.cip_dispatcher = cip_dispatcher
        self.revision = CIPAttribute(class_id=self.class_id, instance_id=b'\x00',
                                     attribute_id=b'\x01', data_type=CIPUnsignedInteger())
        self.interface_configuration_status = CIPAttribute(class_id=self.class_id, instance_id=b'\x01',
                                                           attribute_id=b'\x01', data_type=CIPDoubleWord())
        self.configuration_capability = CIPAttribute(class_id=self.class_id, instance_id=b'\x01',
                                                     attribute_id=b'\x02', data_type=CIPDoubleWord())
        self.configuration_control = CIPAttribute(class_id=self.class_id, instance_id=b'\x01',
                                                  attribute_id=b'\x03', data_type=CIPDoubleWord(), writeable=True)
        physical_link_object_structure = CIPStructure()
        physical_link_object_structure.variable_name = "Physical Link Object"
        physical_link_object_structure.add_member('Path Size', CIPUnsignedInteger())
        # This is a hack! I don't anything about an EPATH type, but it's four bytes with an alignment of 1
        physical_link_object_structure.add_member('Path1', CIPByte())
        physical_link_object_structure.add_member('Path2', CIPByte())
        physical_link_object_structure.add_member('Path3', CIPByte())
        physical_link_object_structure.add_member('Path4', CIPByte())
        self.physical_link_object = CIPAttribute(class_id=self.class_id, instance_id=b'\x01',
                                                 attribute_id=b'\x04', data_type=physical_link_object_structure)
        interface_configuration = CIPStructure()
        interface_configuration.variable_name = "Interface Configuration"
        interface_configuration.add_member('IP Address', CIPUnsignedDoubleInteger())
        interface_configuration.add_member('Subnet Mask', CIPUnsignedDoubleInteger())
        interface_configuration.add_member('Default Gateway', CIPUnsignedDoubleInteger())
        interface_configuration.add_member('Primary Nameserver', CIPUnsignedDoubleInteger())
        interface_configuration.add_member('Secondary Nameserver', CIPUnsignedDoubleInteger())
        interface_configuration.add_member('Domain Name', CIPWord())
        self.interface_configuration = CIPAttribute(class_id=self.class_id, instance_id=b'\x01',
                                                    attribute_id=b'\x05', data_type=interface_configuration,
                                                    writeable=True)
        self.host_name = CIPAttribute(class_id=self.class_id, instance_id=b'\x01',
                                      attribute_id=b'\x06', data_type=CIPDoubleWord(), writeable=True)
        self.encapsulation_inactivity_timeout = CIPAttribute(class_id=self.class_id, instance_id=b'\x01',
                                                             attribute_id=b'\x0d', data_type=CIPDoubleWord(),
                                                             writeable=True)

    def get_attribute_all(self):
        request_path = address_request_path_segment(class_id=self.class_id, instance_id=self.instance_id)
        reply = self.get_attribute_all_from_path(request_path)
        return reply



__author__ = 'Joseph Ryan'
__license__ = "GPLv2"
__maintainer__ = "Joseph Ryan"
__email__ = "jr@aphyt.com"

import socket
import time
from typing import List, Tuple
from aphyt.cip import *


class DataAndAddressItem:
    """
    Data and address items store the data used in common packet format
    EIP-CIP-V2-1.0.pdf Table 2-7.2 – Data and Address item format
    """
    NULL_ADDRESS_ITEM = b'\x00\x00'
    CONNECTED_TRANSPORT_PACKET = b'\xb1\x00'
    UNCONNECTED_MESSAGE = b'\xb2\x00'
    LIST_SERVICES_RESPONSE = b'\x00\x01'
    SOCKADDR_INFO_ORIGINATOR_TO_TARGET = b'\x00\x80'
    SOCKADDR_INFO_TARGET_TO_ORIGINATOR = b'\x01\x80'
    SEQUENCED_ADDRESS_ITEM = b'\x02\x80'

    def __init__(self, type_id, data: bytes):
        self.type_id = type_id
        self.length = len(data).to_bytes(2, 'little')
        self.data = data

    def from_bytes(self, bytes_data_address_item: bytes):
        self.type_id = bytes_data_address_item[0:2]
        self.length = bytes_data_address_item[2:4]
        self.data = bytes_data_address_item[4:]

    def bytes(self):
        return self.type_id + self.length + self.data


class CommonPacketFormat:
    """
    Common Packet Format stores the Data and Address packets. Used in both requests and replies.
    """

    def __init__(self, packets: List[DataAndAddressItem]):
        self.packets = [DataAndAddressItem(DataAndAddressItem.NULL_ADDRESS_ITEM, b''),
                        DataAndAddressItem(DataAndAddressItem.NULL_ADDRESS_ITEM, b'')]
        if len(packets) < 1:
            pass
        elif len(packets) < 2:
            self.packets[1] = packets[0]
        else:
            self.packets = packets
        self.item_count = len(self.packets)
        self.packet_bytes = b''
        for packet in self.packets:
            self.packet_bytes += packet.bytes()

    def from_bytes(self, bytes_common_packet_format: bytes):
        self.item_count = int.from_bytes(bytes_common_packet_format[0:2], 'little')
        self.packet_bytes = bytes_common_packet_format[2:]
        packet_offset = 0
        packet_index = 0
        while packet_offset < len(self.packet_bytes):
            data_address_item_id = self.packet_bytes[packet_offset: packet_offset + 2]
            data_address_item_length = int.from_bytes(self.packet_bytes[packet_offset + 2: packet_offset + 4], 'little')
            data_address_item_data = self.packet_bytes[packet_offset + 4: packet_offset + data_address_item_length + 4]
            self.packets[packet_index] = DataAndAddressItem(data_address_item_id, data_address_item_data)
            packet_offset = packet_offset + data_address_item_length + 4
            packet_index = packet_index + 1

    def bytes(self):
        return self.item_count.to_bytes(2, 'little') + self.packet_bytes


class CommandSpecificData:
    """
    Command Specific Data is sent in an EIP send rr command
    """

    def __init__(self, interface_handle: bytes = b'\x00\x00\x00\x00',
                 timeout: bytes = b'\x08\x00',
                 encapsulated_packet: bytes = b''):
        self.interface_handle = interface_handle
        self.timeout = timeout
        self.encapsulated_packet = encapsulated_packet

    def from_bytes(self, bytes_command_specific_data: bytes):
        self.interface_handle = bytes_command_specific_data[0:4]
        self.timeout = bytes_command_specific_data[4:6]
        self.encapsulated_packet = bytes_command_specific_data[6:]

    def bytes(self):
        return self.interface_handle + self.timeout + self.encapsulated_packet


class EIPMessage:
    """
    Class for creating or parsing Ethernet/IP messages

    EIP Message::
        |-Command Specific Data
          |-Common Packet Format
            |-Data And Address Item
              |-CIP Message
                |-Route Path
    """

    def __init__(self, command=b'\x00\x00', command_data=b'', session_handle_id=b'\x00\x00\x00\x00',
                 status=b'\x00\x00\x00\x00', sender_context_data=b'\x00\x00\x00\x00\x00\x00\x00\x00',
                 command_options=b'\x00\x00\x00\x00'):
        """

        :param command:
        :param command_data:
        :param session_handle_id:
        :param status:
        :param sender_context_data:
        :param command_options:
        """
        self.command = command
        self.length = len(command_data).to_bytes(2, 'little')
        self.session_handle_id = session_handle_id
        self.status = status
        self.sender_context_data = sender_context_data
        self.command_options = command_options
        self.command_data = command_data

    def bytes(self) -> bytes:
        return \
            self.command + self.length + self.session_handle_id + self.status + \
            self.sender_context_data + self.command_options + self.command_data

    def from_bytes(self, eip_message_bytes: bytes):
        self.command = eip_message_bytes[0:2]
        self.length = eip_message_bytes[2:4]
        self.session_handle_id = eip_message_bytes[4:8]
        self.status = eip_message_bytes[8:12]
        self.sender_context_data = eip_message_bytes[12:20]
        self.command_options = eip_message_bytes[20:24]
        self.command_data = eip_message_bytes[24:]


class EIPDispatcher(ABC):
    explicit_message_port = 44818

    def __init__(self):
        super().__init__()
        self.socket = None

    @abstractmethod
    def send_command(self, eip_command: EIPMessage, host: str) -> EIPMessage:
        pass

    def list_identity(self, host=''):
        """
        Used by an originator to locate possible targets
        :return:
        """
        eip_message = EIPMessage(b'\x63\x00')
        return self.send_command(eip_message, host).command_data

    def list_services(self, host=''):
        """
        Find which services a target supports
        :return:
        """
        eip_message = EIPMessage(b'\x04\x00')
        return self.send_command(eip_message, host).command_data

    def list_interfaces(self, host=''):
        """
        Used by an originator to identify possible non-CIP interfaces on the target
        :return:
        """
        eip_message = EIPMessage(b'\x64\x00')
        return self.send_command(eip_message, host).command_data


class EIPConnectedCommandMixin(EIPDispatcher):
    def __init__(self):
        super().__init__()
        self.explicit_message_socket = None
        self.session_handle_id = b'\x00\x00\x00\x00\x00\x00\x00\x00'
        self.is_connected_explicit = False
        self.has_session_handle = False
        self.BUFFER_SIZE = 4096
        self.host = None

    def __del__(self):
        self.close_explicit()

    def send_command(self, eip_command: EIPMessage, host) -> EIPMessage:
        """
        Used to send and receive Ethernet/IP messages
        :param host:
        :param eip_command:
        :return:
        """
        received_eip_message = EIPMessage()
        if self.is_connected_explicit:
            self.explicit_message_socket.send(eip_command.bytes())
            received_data = self.explicit_message_socket.recv(self.BUFFER_SIZE)
            received_eip_message.from_bytes(received_data)
        return received_eip_message

    def connect_explicit(self, host):
        """
        Create and explicit Ethernet/IP connection
        :param host:
        """
        try:
            self.explicit_message_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.explicit_message_socket.connect((host, self.explicit_message_port))
            self.host = host
            self.is_connected_explicit = True
        except socket.error as err:
            if self.explicit_message_socket:
                self.explicit_message_socket.close()
                self.is_connected_explicit = False

    def close_explicit(self):
        """
        Close the explicit Ethernet/IP connection
        :return:
        """
        self.is_connected_explicit = False
        self.has_session_handle = False
        self.host = None
        if self.explicit_message_socket:
            self.explicit_message_socket.close()

    def register_session(self, command_data=b'\x01\x00\x00\x00'):
        """
        Used by an originator to establish a session. It is required before sending CIP messages
        :param command_data:
        :return:
        """
        eip_message = EIPMessage(b'\x65\x00', command_data)
        response = self.send_command(eip_message, self.host)
        self.has_session_handle = True
        self.session_handle_id = response.session_handle_id
        return response


class EIPUnconnectedCommandMixin(EIPDispatcher):
    """
    ToDo add broadcast_command to return a List of EIPMessages (List Service to a full domain)
    """
    def __init__(self):
        super().__init__()

    def send_command(self, eip_command: EIPMessage, host: str) -> List[EIPMessage]:
        received_eip_message = EIPMessage()
        response_list = []
        try:
            udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            udp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            udp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            udp_socket.settimeout(5)
            print(self.explicit_message_port)
            udp_socket.bind(('0.0.0.0', self.explicit_message_port))
            udp_socket.sendto(eip_command.bytes(), (host, self.explicit_message_port))
            received_data = udp_socket.recvfrom(1024)[0]
            # while True:
            #     try:
            #         received_data = udp_socket.recvfrom(1024)[0]
            #         # response_list.append(received_eip_message.from_bytes(received_data))
            #     except socket.error as e:
            #         break
            received_eip_message.from_bytes(received_data)
        except socket.error as e:
            print('Failed to create socket')
            print(e)

        return received_eip_message


class EIPConnectedCIPDispatcher(EIPConnectedCommandMixin, CIPDispatcher):
    """
    EIP is an encapsulation protocol for CIP (common industrial protocol) messages
    """

    null_address_item = b'\x00\x00\x00\x00'
    cip_handle = b'\x00\x00\x00\x00'

    def __init__(self):
        super().__init__()

    def execute_cip_command(self, request: CIPRequest) -> CIPReply:
        """
        Implements the abstract method in order to become a concrete CIPDispatcher class
        :param request:
        :return:
        """
        data_address_item = DataAndAddressItem(DataAndAddressItem.UNCONNECTED_MESSAGE, request.bytes)
        packets = [data_address_item]
        # ToDo add interface handle to track responses?
        common_packet_format = CommonPacketFormat(packets)
        command_specific_data = CommandSpecificData(encapsulated_packet=common_packet_format.bytes())
        response = self.send_rr_data(command_specific_data.bytes()).packets[1].bytes()
        reply_data_and_address_item = DataAndAddressItem('', b'')
        # ToDo is this removing error data?
        reply_data_and_address_item.from_bytes(response)
        cip_reply = CIPReply(reply_data_and_address_item.data)
        return cip_reply

    # @staticmethod
    # def command_specific_data_from_eip_message_bytes(eip_message_bytes: bytes):
    #     """
    #     Extracts command specific data from Ethernet/IP reply bytes
    #     :param eip_message_bytes:
    #     :return:
    #     """
    #     eip_message = EIPMessage()
    #     eip_message.from_bytes(eip_message_bytes)
    #     return EIPConnectedCIPDispatcher.command_specific_data_from_eip_message(eip_message)
    #
    # @staticmethod
    # def command_specific_data_from_eip_message(eip_message: EIPMessage) -> CommandSpecificData:
    #     """
    #     Extracts the command specific data from an Ethernet/IP message
    #     :param eip_message:
    #     :return:
    #     """
    #     command_specific_data = CommandSpecificData()
    #     command_specific_data.from_bytes(eip_message.bytes())
    #     return command_specific_data

    def send_rr_data(self, command_specific_data: bytes) -> CommonPacketFormat:
        """
        Ethernet/IP command to send an encapsulated request and reply packet between originator and target
        :param command_specific_data:
        :return:
        """
        eip_message = EIPMessage(b'\x6f\x00', command_specific_data, self.session_handle_id)
        reply = self.send_command(eip_message, self.host)
        reply_command_specific_data = CommandSpecificData()
        reply_command_specific_data.from_bytes(reply.command_data)
        reply_packet = CommonPacketFormat([])
        reply_packet.from_bytes(reply_command_specific_data.encapsulated_packet)
        return reply_packet

    # def get_attribute_all_service(self, tag_request_path):
    #     get_attribute_all_request = CIPRequest(CIPService.GET_ATTRIBUTE_ALL, tag_request_path)
    #     return self.execute_cip_command(get_attribute_all_request)
    #
    # def get_attribute_single_service(self, tag_request_path):
    #     get_attribute_single_request = CIPRequest(CIPService.GET_ATTRIBUTE_SINGLE, tag_request_path)
    #     return self.execute_cip_command(get_attribute_single_request)



__author__ = 'Joseph Ryan'
__license__ = "GPLv2"
__maintainer__ = "Joseph Ryan"
__email__ = "jr@aphyt.com"

from .cip import *
from .cip_attributes import *
from .cip_datatypes import *


class CIPService:
    """
    ToDo eventually all services should be mixins and then the cip dispatcher will not contain the
    reading, writing, getting, setting and resetting methods
    """
    READ_TAG_SERVICE = b'\x4c'
    READ_TAG_FRAGMENTED_SERVICE = b'\x52'
    WRITE_TAG_SERVICE = b'\x4d'
    WRITE_TAG_FRAGMENTED_SERVICE = b'\x53'
    READ_MODIFY_WRITE_TAG_SERVICE = b'\x4e'
    # w506_nx_nj - series_cpu_unit_built - in_ethernet_ip_port_users_manual_en.pdf
    # 303 of 570 CIP Object Services
    GET_ATTRIBUTE_ALL = b'\x01'
    GET_ATTRIBUTE_SINGLE = b'\x0e'
    RESET = b'\x05'
    SET_ATTRIBUTE_SINGLE = b'\x10'

    def __init__(self, cip_dispatcher: CIPDispatcher, **kwargs):
        self.cip_dispatcher = cip_dispatcher


class ReadTagServiceMixin(CIPService):
    def __init__(self, cip_dispatcher: CIPDispatcher):
        super().__init__(cip_dispatcher)

    def read_tag(self, tag_request_path, number_of_elements=1) -> CIPReply:
        service_code = b'\x4c'
        read_tag_request = \
            CIPRequest(service_code, tag_request_path, number_of_elements.to_bytes(2, 'little'))
        return self.cip_dispatcher.execute_cip_command(read_tag_request)


class WriteTagServiceMixin(CIPService):
    def __init__(self, cip_dispatcher: CIPDispatcher):
        super().__init__(cip_dispatcher)

    def write_tag(self, tag_request_path, request_service_data: CIPCommonFormat, number_of_elements=1):
        service_code = b'\x4d'
        data = request_service_data.data_type + \
            int(request_service_data.additional_info_length).to_bytes(1, 'little') + \
            request_service_data.additional_info + number_of_elements.to_bytes(2, 'little') + \
            request_service_data.data
        write_tag_request = CIPRequest(service_code, tag_request_path, data)
        return self.cip_dispatcher.execute_cip_command(write_tag_request)


class GetAttributeAllMixin(CIPService):
    def __init__(self, cip_dispatcher: CIPDispatcher):
        super().__init__(cip_dispatcher)

    def get_attribute_all_from_path(self, tag_request_path):
        service_code = b'\x01'
        get_attribute_all_request = CIPRequest(service_code, tag_request_path)
        return self.cip_dispatcher.execute_cip_command(get_attribute_all_request)

    # def get_attribute_all(self, cip_attribute: CIPAttribute) -> bytes:
    #     reply = self.get_attribute_all_from_path(cip_attribute.request_path)
    #     return reply.reply_data


class GetAttributeSingleMixin(CIPService):
    """Always use the Set mixins before the get mixins so that """
    def __init__(self, cip_dispatcher: CIPDispatcher, **kwargs):
        super().__init__(cip_dispatcher, **kwargs)

    def get_attribute_single_from_path(self, tag_request_path) -> CIPReply:
        service_code = b'\x0e'
        get_attribute_single_request = \
            CIPRequest(service_code, tag_request_path)
        return self.cip_dispatcher.execute_cip_command(get_attribute_single_request)

    def get_attribute_single(self, cip_attribute: CIPAttribute) -> CIPAttribute:
        reply = self.get_attribute_single_from_path(cip_attribute.request_path)
        cip_attribute.data_type.data = reply.reply_data
        return cip_attribute


class SetAttributeSingleMixin(CIPService):
    def __init__(self, cip_dispatcher: CIPDispatcher, **kwargs):
        super().__init__(cip_dispatcher, **kwargs)

    def set_attribute_single_from_path(self, tag_request_path, data: bytes):
        service_code = b'\x10'
        set_attribute_single_request = CIPRequest(service_code, tag_request_path, data)
        return self.cip_dispatcher.execute_cip_command(set_attribute_single_request)

    def set_attribute_single(self, cip_attribute: CIPAttribute):
        assert (cip_attribute.writeable is True)
        self.set_attribute_single_from_path(cip_attribute.request_path, data=cip_attribute.data_type.data)


class ResetMixin(CIPService):
    def __init__(self, cip_dispatcher: CIPDispatcher):
        super().__init__(cip_dispatcher)

    def reset(self, tag_request_path):
        service_code = b'\x05'
        reset_request = CIPRequest(service_code, tag_request_path)
        return self.cip_dispatcher.execute_cip_command(reset_request)



__author__ = 'Joseph Ryan'
__license__ = "GPLv2"
__maintainer__ = "Joseph Ryan"
__email__ = "jr@aphyt.com"

from aphyt.eip import *


class VariableTypeObjectReply(CIPReply):
    """
    CIP Reply from the Get Attribute All service to Variable Type Object Class Code 0x6C adding descriptive properties
    Omron Vendor specific
    """

    def __init__(self, reply_bytes: bytes):
        super().__init__(reply_bytes=reply_bytes)

    @property
    def size_in_memory(self):
        return struct.unpack("<L", self.reply_data[0:4])[0]

    @property
    def cip_data_type(self):
        return self.reply_data[5:6]

    @property
    def cip_data_type_of_array(self):
        return self.reply_data[6:7]

    @property
    def array_dimension(self):
        return struct.unpack("<B", self.reply_data[7:8])[0]

    @property
    def number_of_elements(self):
        """Number of elements in each dimension of  the array"""
        dimension_size_list = []
        for i in range(self.array_dimension):
            dimension_size = struct.unpack("<L", self.reply_data[8 + i * 4:8 + i * 4 + 4])[0]
            dimension_size_list.append(dimension_size)
        return dimension_size_list

    @property
    def number_of_members(self):
        return struct.unpack("<H", self.reply_data[8 + self.array_dimension * 4:10 + self.array_dimension * 4])[0]

    @property
    def crc_code(self):
        return struct.unpack("<H", self.reply_data[14 + self.array_dimension * 4:16 + self.array_dimension * 4])[0]

    @property
    def variable_type_name_length(self):
        return struct.unpack("<B", self.reply_data[16 + self.array_dimension * 4:17 + self.array_dimension * 4])[0]

    @property
    def padding(self):
        if self.variable_type_name_length % 2 == 0:
            return 1
        else:
            return 0

    @property
    def variable_type_name(self):
        return \
            self.reply_data[17 + self.array_dimension * 4:
                            17 + self.array_dimension * 4 + self.variable_type_name_length]

    @property
    def next_instance_id(self):
        return \
            self.reply_data[self.padding + 17 + self.array_dimension * 4 + self.variable_type_name_length:
                            self.padding + 17 + self.array_dimension * 4 + self.variable_type_name_length + 4]

    @property
    def nesting_variable_type_instance_id(self):
        return \
            self.reply_data[self.padding + 21 + self.array_dimension * 4 + self.variable_type_name_length:
                            self.padding + 21 + self.array_dimension * 4 + self.variable_type_name_length + 4]

    @property
    def start_array_elements(self):
        """Number of elements in each dimension of  the array"""
        array_start_list = []
        for i in range(self.array_dimension):
            array_start = \
                struct.unpack("<L",
                              self.reply_data[self.padding + 25 +
                                              self.array_dimension * 4 +
                                              self.variable_type_name_length:
                                              self.padding + 25 +
                                              self.array_dimension * 4 +
                                              self.variable_type_name_length + 4])[0]
            array_start_list.append(array_start)
        return array_start_list


class VariableObjectReply(CIPReply):
    """
    CIP Reply from the Get Attribute All service to Variable Object Class Code 0x6B adding descriptive properties
    Omron Vendor specific
    """

    def __init__(self, reply_bytes: bytes):
        super().__init__(reply_bytes=reply_bytes)

    @property
    def size(self):
        return struct.unpack("<L", self.reply_data[0:4])[0]

    @property
    def cip_data_type(self):
        return self.reply_data[4:5]

    @property
    def cip_data_type_of_array(self):
        return self.reply_data[5:6]

    @property
    def array_dimension(self):
        # One byte of padding after this. Skip to byte 8
        return struct.unpack("<B", self.reply_data[6:7])[0]

    @property
    def number_of_elements(self):
        """Number of elements in each dimension of  the array"""
        dimension_size_list = []
        for i in range(self.array_dimension):
            dimension_size = struct.unpack("<L", self.reply_data[8 + i * 4:8 + i * 4 + 4])[0]
            dimension_size_list.append(dimension_size)
        return dimension_size_list

    @property
    def bit_number(self):
        return struct.unpack("<B", self.reply_data[16 + self.array_dimension * 4:17 + self.array_dimension * 4])[0]

    @property
    def variable_type_instance_id(self):
        return self.reply_data[20 + self.array_dimension * 4:24 + self.array_dimension * 4]

    @property
    def start_array_elements(self):
        """Number of elements in each dimension of  the array"""
        array_start_list = []
        for i in range(self.array_dimension):
            array_start = \
                struct.unpack("<L",
                              self.reply_data[24 + self.array_dimension * 4:
                                              24 + self.array_dimension * 4 + 4])[0]
            array_start_list.append(array_start)
        return array_start_list


class SimpleDataSegmentRequest:
    """
    A simple data segment request is the Omron specific format for requesting data that will not fit
    in a single CIP message. This is used for strings, arrays and structures
    """
    def __init__(self, offset, size):
        self.simple_data_type_code = b'\x80'
        self.segment_length = b'\x03'  # Fixed and in words
        self.offset = offset
        self.size = size

    def bytes(self):
        return \
            self.simple_data_type_code + self.segment_length + \
            struct.pack("<L", self.offset) + struct.pack("<H", self.size)


class NSeries:
    """
    Concrete implementation of EIP abstract base class that implements Omron specific Ethernet/IP
    services in addition to the common Ethernet/IP services of the parent class and the CIP specific
    items in the grandparent class
    """
    MAXIMUM_LENGTH = 502  # UCMM maximum length is 502 bytes

    def __init__(self):
        super().__init__()
        self.derived_data_type_dictionary = {}
        self.connected_cip_dispatcher = EIPConnectedCIPDispatcher()

    def connect_explicit(self, host):
        self.connected_cip_dispatcher.connect_explicit(host)

    def close_explicit(self):
        self.connected_cip_dispatcher.close_explicit()

    def register_session(self):
        self.connected_cip_dispatcher.register_session()

    def get_instance_list_service(self, tag_request_path: bytes, data: bytes):
        """
        Omron specific CIP service
        :param tag_request_path:
        :param data:
        :return:
        """
        get_instance_list_request = CIPRequest(b'\x5f', tag_request_path, data)
        return self.connected_cip_dispatcher.execute_cip_command(get_instance_list_request)

    def update_derived_data_type_dictionary(self):
        # ToDo get the derived data types in such  a way they are easy to use
        number_of_entries = self._get_number_of_derived_data_types()
        for index in range(1, number_of_entries+1):
            request_path = eip.address_request_path_segment(class_id=b'\x6c', instance_id=index.to_bytes(2, 'little'))
            reply = VariableTypeObjectReply(
                self.connected_cip_dispatcher.get_attribute_all_service(request_path).bytes)
            print('index: %s '
                  'size: %s '
                  'type: %s '
                  'array type: %s '
                  'dimensions: %s '
                  'elements: %s '
                  'number of members: %s '
                  'crc: %s '
                  'variable name length: %s '
                  'variable type name: %s '
                  'next instance id: %s '
                  'nesting variable type instance id %s '
                  'start array elements %s' %
                  (str(index),
                   str(reply.size_in_memory),
                   str(reply.cip_data_type),
                   str(reply.cip_data_type_of_array),
                   str(reply.array_dimension),
                   str(reply.number_of_elements),
                   str(reply.number_of_members),
                   str(reply.crc_code),
                   str(reply.variable_type_name_length),
                   str(reply.variable_type_name),
                   str(reply.next_instance_id),
                   str(reply.nesting_variable_type_instance_id),
                   str(reply.start_array_elements)
                   ))

    def update_variable_dictionary(self):
        """
        Make sure the variable dictionary is populated with the
        latest variable and datatype information from controller
        :return:
        """
        update_data_type_dictionary(self.connected_cip_dispatcher.data_type_dictionary)
        variable_list = self._get_variable_list()
        instance_id = 1
        for variable in variable_list:
            reply = self._get_variable_object(instance_id)
            variable_cip_datatype = self.connected_cip_dispatcher.data_type_dictionary.get(reply.cip_data_type)()
            if not isinstance(variable_cip_datatype, type(None)):
                variable_cip_datatype.instance_id = instance_id
                # Instantiate the classes into objects
                variable_cip_datatype = self._get_data_instance(variable_cip_datatype)
                variable_cip_datatype.variable_name = str(variable)
                self.connected_cip_dispatcher.variables.update({variable: variable_cip_datatype})
                if variable[0:1] == '_':
                    self.connected_cip_dispatcher.system_variables.update({variable: variable_cip_datatype})
                else:
                    self.connected_cip_dispatcher.user_variables.update({variable: variable_cip_datatype})
            instance_id = instance_id + 1

    def _get_data_instance(self, cip_datatype_instance: CIPDataType) -> CIPDataType:
        """
        This method is to get an instance of a CIP data type from its type definition.
        :param cip_datatype_instance:
        :return:
        """
        # ToDo TESTS
        if isinstance(cip_datatype_instance, CIPArray):
            variable_object_reply = self._get_variable_object(cip_datatype_instance.instance_id)
            if variable_object_reply.variable_type_instance_id == b'\x00\x00\x00\x00':
                cip_datatype_instance.from_items(variable_object_reply.cip_data_type_of_array,
                                                 variable_object_reply.size,
                                                 variable_object_reply.array_dimension,
                                                 variable_object_reply.number_of_elements,
                                                 variable_object_reply.start_array_elements)
            else:
                member_instance_id = int.from_bytes(variable_object_reply.variable_type_instance_id, 'little')
                member_instance = self._get_member_instance(member_instance_id)
                cip_datatype_instance.from_instance(member_instance,
                                                    variable_object_reply.size,
                                                    variable_object_reply.array_dimension,
                                                    variable_object_reply.number_of_elements,
                                                    variable_object_reply.start_array_elements)
            return cip_datatype_instance

        elif isinstance(cip_datatype_instance, CIPString):
            variable_object_reply = self._get_variable_object(cip_datatype_instance.instance_id)
            cip_datatype_instance.size = variable_object_reply.size
            return cip_datatype_instance

        elif isinstance(cip_datatype_instance, CIPAbbreviatedStructure):
            pass

        elif isinstance(cip_datatype_instance, CIPStructure):
            variable_object_reply = self._get_variable_object(cip_datatype_instance.instance_id)
            variable_type_instance_id = int.from_bytes(variable_object_reply.variable_type_instance_id, 'little')
            variable_type_object_reply = self._get_variable_type_object(variable_type_instance_id)
            cip_datatype_instance.variable_type_name = str(variable_type_object_reply.variable_type_name, 'utf-8')
            cip_datatype_instance.size = variable_type_object_reply.size_in_memory
            nesting_variable_type_instance_id = \
                int.from_bytes(variable_type_object_reply.nesting_variable_type_instance_id, 'little')
            member_instance_id = nesting_variable_type_instance_id
            while member_instance_id != 0:
                member_cip_datatype_instance = self._get_member_instance(member_instance_id)
                cip_datatype_instance.members.update(
                    {member_cip_datatype_instance.variable_type_name: member_cip_datatype_instance})
                member_instance_id = \
                    int.from_bytes(member_cip_datatype_instance.next_instance_id, 'little')
            return cip_datatype_instance
        else:
            cip_datatype_instance = self.connected_cip_dispatcher.data_type_dictionary.get(cip_datatype_instance.data_type_code())()
            return cip_datatype_instance

    def _get_member_instance(self, member_instance_id: int) -> CIPDataType:
        """
        This method returns a CIP datatype instance from a member ID. This is how the driver can
        build instances of derived data types like structures and arrays of structures
        :param member_instance_id:
        :return:
        """
        variable_type_object_reply = self._get_variable_type_object(member_instance_id)
        cip_datatype_instance = self.connected_cip_dispatcher.data_type_dictionary.get(variable_type_object_reply.cip_data_type)()
        cip_datatype_instance.variable_type_name = str(variable_type_object_reply.variable_type_name, 'utf-8')
        cip_datatype_instance.size = variable_type_object_reply.size_in_memory
        cip_datatype_instance.next_instance_id = variable_type_object_reply.next_instance_id
        cip_datatype_instance.nesting_variable_type_instance_id = \
            variable_type_object_reply.nesting_variable_type_instance_id
        if isinstance(cip_datatype_instance, CIPArray):
            array_start_list = []
            for i in range(variable_type_object_reply.array_dimension):
                array_start = 0
                array_start_list.append(array_start)
            cip_datatype_instance.from_items(variable_type_object_reply.cip_data_type_of_array,
                                             variable_type_object_reply.size_in_memory,
                                             variable_type_object_reply.array_dimension,
                                             variable_type_object_reply.number_of_elements,
                                             array_start_list)
            return cip_datatype_instance

        elif isinstance(cip_datatype_instance, CIPString):
            cip_datatype_instance.size = variable_type_object_reply.size_in_memory
            return cip_datatype_instance

        elif isinstance(cip_datatype_instance, CIPAbbreviatedStructure):
            pass

        elif isinstance(cip_datatype_instance, CIPStructure):
            member_instance_id = int.from_bytes(cip_datatype_instance.nesting_variable_type_instance_id, 'little')
            while member_instance_id != 0:
                variable_type_object_reply = self._get_variable_type_object(member_instance_id)
                member_cip_datatype_instance = self._get_member_instance(member_instance_id)
                if member_cip_datatype_instance.alignment > cip_datatype_instance.alignment:
                    cip_datatype_instance.alignment = member_cip_datatype_instance.alignment
                cip_datatype_instance.members.update(
                    {member_cip_datatype_instance.variable_type_name: member_cip_datatype_instance})
                member_instance_id = \
                    int.from_bytes(variable_type_object_reply.next_instance_id, 'little')
            return cip_datatype_instance
        else:
            # cip_datatype_instance = self.data_type_dictionary.get(cip_datatype_instance.data_type_code())()
            return cip_datatype_instance

    def read_variable(self, variable_name: str):
        """
        This method will read the variable name from the controller and return it in the corresponding
        Python datatype
        :param variable_name:
        :return:
        """
        request_path = variable_request_path_segment(variable_name)
        cip_datatype_object = self.connected_cip_dispatcher.variables.get(variable_name)
        if isinstance(cip_datatype_object, (CIPString, CIPArray, CIPStructure, CIPAbbreviatedStructure)):
            return self._multi_message_variable_read(cip_datatype_object)
        else:
            response = self.connected_cip_dispatcher.read_tag_service(request_path)
            cip_datatype_object.from_bytes(response.reply_data)
            return cip_datatype_object.value()

    def write_variable(self, variable_name: str, data):
        """
        This method takes a variable name and formats the Python datatype into the correct CIP datatype
        and writes it to the controller
        :param variable_name:
        :param data:
        :return:
        """
        request_path = variable_request_path_segment(variable_name)
        cip_datatype_object = self.connected_cip_dispatcher.variables.get(variable_name)
        cip_datatype_object.from_value(data)
        if isinstance(cip_datatype_object, (CIPString, CIPArray, CIPStructure, CIPAbbreviatedStructure)):
            self._multi_message_variable_write(cip_datatype_object, data)
        else:
            request_data = CIPCommonFormat(cip_datatype_object.data_type_code(), data=cip_datatype_object.data)
            self.connected_cip_dispatcher.write_tag_service(request_path, request_data)

    def _multi_message_variable_read(self, cip_datatype_object: CIPDataType, offset=0):
        """
        This method is to read data that does not fit into a single CIP message
        :param cip_datatype_object:
        :param offset:
        :return:
        """
        max_read_size = self.MAXIMUM_LENGTH - 8
        data = b''
        while offset < cip_datatype_object.size:
            if cip_datatype_object.size - offset > max_read_size:
                read_size = max_read_size
            else:
                read_size = cip_datatype_object.size - offset
            response = self._simple_data_segment_read(cip_datatype_object, offset, read_size)
            reply_bytes = response.reply_data
            cip_common_format = CIPCommonFormat()
            cip_common_format.from_bytes(reply_bytes)
            if isinstance(cip_datatype_object, CIPString):
                # First two characters of the string seem to be how many characters were read
                data = data + cip_common_format.data[2:]
            elif isinstance(cip_datatype_object, CIPStructure):
                cip_datatype_object.crc_code = cip_common_format.additional_info
                data = data + cip_common_format.data
            else:
                data = data + cip_common_format.data
            offset = offset + max_read_size
        cip_datatype_object.data = data
        # cip_datatype_object.size = len(data) # Removed Why did it exist? If weird stuff breaks revisit
        return cip_datatype_object.value()

    def _multi_message_variable_write(self, cip_datatype_object: CIPDataType, data, offset=0):
        """
        This method is to write data that does not fit into a single CIP message
        :param cip_datatype_object:
        :param data:
        :param offset:
        :return:
        """
        max_write_size = 400
        cip_datatype_object.from_value(data)
        if isinstance(cip_datatype_object, CIPArray):
            max_write_size = max_write_size // cip_datatype_object.array_data_type_size * \
                             cip_datatype_object.array_data_type_size
        while offset < cip_datatype_object.size:
            if cip_datatype_object.size - offset > max_write_size:
                write_size = max_write_size
            else:
                write_size = cip_datatype_object.size - offset
            response = self._simple_data_segment_write(
                cip_datatype_object, offset, write_size,
                cip_datatype_object.data[offset:offset + write_size])
            offset = offset + max_write_size

    def _simple_data_segment_read(self, cip_datatype_object: CIPDataType, offset, read_size):
        """
        This method formats as request path to be used with simple_data_segment reading operations
        :param cip_datatype_object:
        :param offset:
        :param read_size:
        :return:
        """
        # ToDo consider the request path should reside with the variable (possibly converted to logical segment)
        request_path = variable_request_path_segment(cip_datatype_object.variable_name)
        simple_data_request_path = SimpleDataSegmentRequest(offset, read_size)
        request_path = request_path + simple_data_request_path.bytes()
        response = self.connected_cip_dispatcher.read_tag_service(request_path)
        return response

    def _simple_data_segment_write(self, cip_datatype_object: CIPDataType, offset, write_size, data):
        """
        This method formats as request path to be used with simple_data_segment writing operations
        :param cip_datatype_object:
        :param offset:
        :param write_size:
        :param data:
        :return:
        """
        request_path = variable_request_path_segment(cip_datatype_object.variable_name)
        simple_data_request_path = SimpleDataSegmentRequest(offset, write_size)
        request_path = request_path + simple_data_request_path.bytes()
        # ToDo how to prevent testing type code here. Function is doing too much
        response = None
        if cip_datatype_object.data_type_code() == CIPString.data_type_code():
            data = struct.pack("<H", len(data)) + data
            request_data = CIPCommonFormat(cip_datatype_object.data_type_code(), data=data)
            response = self.connected_cip_dispatcher.write_tag_service(request_path, request_data)
        elif cip_datatype_object.data_type_code() == CIPArray.data_type_code():
            # ToDo Test String array. Probably have to put the length
            if cip_datatype_object.array_data_type == CIPStructure.data_type_code():
                array_variable_object = self._get_variable_object(cip_datatype_object.instance_id)
                structure_variable_type_object = \
                    self._get_variable_type_object(
                        int.from_bytes(array_variable_object.variable_type_instance_id, 'little'))
                crc_code = structure_variable_type_object.crc_code.to_bytes(2, 'little')
                request_data = CIPCommonFormat(CIPAbbreviatedStructure.data_type_code(), additional_info_length=2,
                                               additional_info=crc_code,
                                               data=data)
            else:
                request_data = CIPCommonFormat(cip_datatype_object.array_data_type, data=data)
            response = self.connected_cip_dispatcher.write_tag_service(request_path, request_data)
        elif cip_datatype_object.data_type_code() == CIPStructure.data_type_code():
            request_data = CIPCommonFormat(CIPAbbreviatedStructure.data_type_code(), additional_info_length=2,
                                           additional_info=cip_datatype_object.crc_code, data=data)
            response = self.connected_cip_dispatcher.write_tag_service(request_path, request_data)
        return response

    def _get_variable_type_object(self, instance_id: int) -> VariableTypeObjectReply:
        """
        Omron specific CIP class that is used to describe variable types. This is where derived data types
        will have their member definitions
        :param instance_id:
        :return:
        """
        request_path = address_request_path_segment(
            class_id=b'\x6c', instance_id=instance_id.to_bytes(2, 'little'))
        variable_type_object_reply = VariableTypeObjectReply(self.connected_cip_dispatcher.get_attribute_all_service(request_path).bytes)
        return variable_type_object_reply

    def _get_number_of_derived_data_types(self) -> int:
        request_path = eip.address_request_path_segment(class_id=b'\x6c', instance_id=int(0).to_bytes(2, 'little'))
        reply = self.connected_cip_dispatcher.get_attribute_all_service(request_path)
        max_instance = struct.unpack("<H", reply.reply_data[2:4])[0]
        return max_instance

    def _get_variable_object(self, instance_id: int) -> VariableObjectReply:
        """
        Omron specific CIP class that is used to describe variables. This contains critical information if the
        variable is an array type
        :param instance_id:
        :return:
        """
        request_path = address_request_path_segment(
            class_id=b'\x6b', instance_id=instance_id.to_bytes(2, 'little'))
        variable_object_reply = VariableObjectReply(self.connected_cip_dispatcher.get_attribute_all_service(request_path).bytes)
        return variable_object_reply

    def _get_variable_list(self):
        """
        Omron specific method for creating a list of variables that are published in the controller.
        :return:
        """
        tag_list = []
        for tag_index in range(self._get_number_of_variables()):
            offset = tag_index + 1
            request_path = address_request_path_segment(b'\x6a', offset.to_bytes(2, 'little'))
            reply = self.connected_cip_dispatcher.get_attribute_all_service(request_path)
            tag = str(reply.reply_data[5:5 + int.from_bytes(reply.reply_data[4:5], 'little')], 'utf-8')
            tag_list.append(tag)
        return tag_list

    def _get_number_of_variables(self) -> int:
        """
        Omron specific method to find number of variables from Tag Name Server
        :return:
        """
        request_path = address_request_path_segment(b'\x6a', b'\x00\x00')
        reply = self.connected_cip_dispatcher.get_attribute_all_service(request_path)
        return int.from_bytes(reply.reply_data[2:4], 'little')

    def _cip_string_read(self, request_path):
        pass

    def _cip_string_write(self):
        pass

    def _cip_structure_read(self):
        pass

    def _cip_structure_write(self):
        pass

    def _cip_abbreviated_structure_read(self):
        pass

    def _cip_abbreviated_structure_write(self):
        pass

    def _cip_array_read(self):
        pass

    def _cip_array_write(self):
        pass



__author__ = 'Joseph Ryan'
__license__ = "GPLv2"
__maintainer__ = "Joseph Ryan"
__email__ = "jr@aphyt.com"


from aphyt.cip.cip_datatypes import *
from aphyt.cip.cip import *


class CIPObject:
    def __init__(self, class_id: bytes, **kwargs):
        self.class_id = class_id


class CIPAttribute:
    def __init__(self,
                 class_id: bytes,
                 instance_id: bytes,
                 attribute_id: bytes,
                 data_type: CIPDataType,
                 writeable: bool = False):
        self.class_id = class_id
        self.instance_id = instance_id
        self.attribute_id = attribute_id
        self.data_type = data_type
        self.writeable = writeable
        self.request_path = address_request_path_segment(self.class_id,
                                                         self.instance_id,
                                                         self.attribute_id)

    def __repr__(self):
        return str(self.data_type.value())



__author__ = 'Joseph Ryan'
__license__ = "GPLv2"
__maintainer__ = "Joseph Ryan"
__email__ = "jr@aphyt.com"

import time
import tkinter
from tkinter import *
from tkinter import ttk
from aphyt.tkmi import helper
from aphyt.eip import EIPConnectedCIPDispatcher
from aphyt.omron import k6pm_th
import concurrent.futures
import threading
import queue
import copy


class EIPThreadDispatcher:
    def __init__(self, instance: EIPConnectedCIPDispatcher):
        self.instance = instance
        self.message_queue = queue.Queue()
        self.message_timeout = 0.5
        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
        self.futures = []
        self.status_integer = 0
        self.services = None

    def start_keep_alive(self):
        if self.instance.is_connected_explicit:
            future = self.executor.submit(
                self.instance.list_services, '')
            self.services = future.result()
            delay = threading.Timer(0.05, self.start_keep_alive)
            delay.start()


class Segment(ttk.Frame):
    def __init__(self, master):
        super().__init__(master)
        # Get the current screen width and height
        screen_width = self.winfo_screenwidth()
        screen_height = self.winfo_screenheight()
        self.unique_id = str(self)
        if screen_height > screen_width:
            max_dimension = screen_height
        else:
            max_dimension = screen_width
        self.segment_color = ttk.Style(self)
        self.theme_name = self.unique_id + '.TFrame'
        self.segment_color.configure(self.theme_name, background='blue')
        self.configure(width=max_dimension * .025, height=max_dimension * .025, style=self.theme_name)

    def set_color(self, color: str):
        self.segment_color.configure(self.theme_name, background=color)
        # self.configure(background=color)


class SegmentViewer(ttk.Frame):
    def __init__(self, master):
        super().__init__(master)
        self.box_grid = [None] * 16
        for i in range(16):
            row = i // 4 + 1
            column = i % 4 + 1
            self.box_grid[i] = Segment(self)
            self.box_grid[i].grid(row=row, column=column)
        for child in self.winfo_children():
            child.grid_configure(padx=1, pady=1)


class TemperaturePixel(ttk.Frame):
    def __init__(self, master):
        super().__init__(master)
        # Get the current screen width and height
        screen_width = self.winfo_screenwidth()
        screen_height = self.winfo_screenheight()
        self.unique_id = str(self)
        if screen_height > screen_width:
            max_dimension = screen_height
        else:
            max_dimension = screen_width
        self.segment_color = ttk.Style(self)
        self.theme_name = self.unique_id + '.TFrame'
        self.segment_color.configure(self.theme_name, background='blue')
        self.configure(width=max_dimension * .0025, height=max_dimension * .0025, style=self.theme_name)

    def set_color(self, color: str):
        self.segment_color.configure(self.theme_name, background=color)


class TemperaturePixelViewer(ttk.Frame):
    def __init__(self, master):
        super().__init__(master)
        # self.box_grid = [None] * 1024
        self.box_grid = []
        for i in range(1024):
            row = i // 32 + 1
            column = i % 32 + 1
            self.box_grid.append(TemperaturePixel(self))
            self.box_grid[i].grid(row=row, column=column, padx=1, pady=1)


class TemperatureCanvas(ttk.Frame):
    def __init__(self, master):
        super().__init__(master)
        screen_width = master.winfo_screenwidth()
        square = 0.25
        self.width = screen_width * square
        self.canvas = tkinter.Canvas(master, width=self.width, height=self.width, bg="white")
        self.next_canvas = tkinter.Canvas(master, width=self.width, height=self.width, bg="white")
        self.pad = 0.5
        self.pixel_size = self.width/32
        self.canvas.configure(highlightthickness=0, borderwidth=0)
        self.next_canvas.configure(highlightthickness=0, borderwidth=0)
        self.canvas.grid(column=1, row=1, sticky=(E, W))
        self.next_canvas.grid(column=1, row=1, sticky=(E, W))
        self.rectangles = []
        self.next_rectangles = []
        color_string = 'green'
        for i in range(1024):
            pos_x = i % 32
            pos_y = i // 32
            x1 = pos_x * self.pixel_size + self.pad
            y1 = pos_y * self.pixel_size + self.pad
            x2 = x1 + self.pixel_size - 2 * self.pad
            y2 = y1 + self.pixel_size - 2 * self.pad
            self.rectangles.append(self.canvas.create_rectangle(x1, y1, x2, y2, width=0, fill=color_string))
            self.next_rectangles.append(self.next_canvas.create_rectangle(x1, y1, x2, y2, width=0, fill=color_string))

        self.next_canvas.grid_remove()
        for child in self.winfo_children():
            child.grid_configure(pady=10, padx=10)

    def set_color_data(self, color_list: list):
        min = 23.0
        max = 31.0
        # print(color_list)
        # self.canvas.delete(tkinter.ALL)
        for i in range(1024):
            pixel_offset = i % 64
            instance_array = i // 64
            pixel_color = color_list[instance_array][pixel_offset]
            pixel_color = k6pm_th.unsigned_integer_to_temp(pixel_color)
            slope = (650.0 - 440.0)/(max-min)
            wavelength = float(pixel_color) * slope + 440.0 - slope * min
            r, g, b = helper.wavelength_to_rgb(wavelength)
            color_string = '#' + '{:02x}'.format(r) + '{:02x}'.format(g) + '{:02x}'.format(b)
            self.next_canvas.itemconfig(self.next_rectangles[i], fill=color_string)
            # self.canvas.create_rectangle(x1, y1, x2, y2, fill=color_string)
        self.canvas, self.next_canvas = self.next_canvas, self.canvas
        self.canvas.grid()
        self.next_canvas.grid_remove()


class ConnectFrame(ttk.Frame):
    def __init__(self, master, eip_thread_dispatcher: EIPThreadDispatcher, connect_function):
        super().__init__(master)
        self.connect_function = connect_function
        self.eip_thread_dispatcher = eip_thread_dispatcher
        self.connect_string = StringVar()
        self.connect_string.set('192.168.250.30')
        self.connect_entry = ttk.Entry(self, width=7, textvariable=self.connect_string)
        self.connect_entry.grid(column=1, row=1, columnspan=2, sticky=(W, E), padx=1)
        self.connect_button = ttk.Button(self, text="Connect", command=self.connect)
        self.connect_button.grid(column=1, row=2, sticky=(W, E))
        self.disconnect_button = ttk.Button(self, text="Disconnect", command=self.disconnect)
        self.disconnect_button.grid(column=2, row=2, sticky=(W, E))
        for child in self.winfo_children():
            child.grid_configure(pady=1)

    def connect(self):
        if not self.eip_thread_dispatcher.instance.is_connected_explicit:
            self.eip_thread_dispatcher.instance.connect_explicit(self.connect_string.get())
            self.eip_thread_dispatcher.instance.register_session()
            self.eip_thread_dispatcher.start_keep_alive()
            self.connect_function()

    def disconnect(self):
        self.eip_thread_dispatcher.executor.submit(
                self.eip_thread_dispatcher.instance.close_explicit)
        # self.eip_thread_dispatcher.instance.close_explicit()


class Screen(Tk):
    def __init__(self):
        super().__init__()
        self.title("K6PM Visualization")
        self.mainframe = ttk.Frame(self, padding="3 3 12 12")
        self.mainframe.grid(column=0, row=0, sticky=(N, W, E, S))
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        self.k6pm_instance = k6pm_th.K6PMTH()
        self.k6pm_thread_dispatcher = EIPThreadDispatcher(self.k6pm_instance.connected_cip_dispatcher)

        self.temp_canvas = TemperatureCanvas(self.mainframe)
        self.temp_canvas.grid(column=1, row=1)
        self.colors = [700] * 64
        self.colors = [self.colors] * 16
        self.temp_canvas.set_color_data(self.colors)

        self.connect_box = ConnectFrame(self.mainframe, self.k6pm_thread_dispatcher, self.periodic_task)
        self.connect_box.grid(column=1, row=2)

        for child in self.mainframe.winfo_children():
            child.grid_configure(padx=5, pady=5)

        self.protocol("WM_DELETE_WINDOW", self.clean_up)

        self.mainloop()

    def calculate(self):
        self.temperature_view.set_color('blue')

    def update_color(self, value):
        # r, g, b = scale_to_rgb(float(value) * 100.0)
        r, g, b = helper.wavelength_to_rgb(float(value) * (650.0-440.0) + 440.0)
        color_string = '#' + '{:02x}'.format(r) + '{:02x}'.format(g) + '{:02x}'.format(b)
        # self.four_pack[1].set_color(color_string)
        self.four_pack.box_grid[0].set_color(color_string)

    def periodic_task(self, period=.5):
        if self.k6pm_thread_dispatcher.instance.is_connected_explicit:
            future = self.k6pm_thread_dispatcher.executor.submit(
                self.k6pm_instance.pixel_temperatures, 1)
            self.colors = future.result()
            self.temp_canvas.set_color_data(self.colors)
            delay = threading.Timer(period, self.periodic_task)
            delay.start()

    def clean_up(self):
        self.connect_box.disconnect()
        self.after_idle(self.destroy)


if __name__ == "__main__":
    window = Screen()



__author__ = 'Joseph Ryan'
__license__ = "GPLv2"
__maintainer__ = "Joseph Ryan"
__email__ = "jr@aphyt.com"

import collections
import struct
import copy
from abc import ABC, abstractmethod
import binascii


def test_bit(int_type, offset):
    mask = 1 << offset
    if int_type & mask != 0:
        response = True
    else:
        response = False
    return response


def set_bit(int_type, offset):
    mask = 1 << offset
    return int_type | mask


def clear_bit(int_type, offset):
    mask = ~(1 << offset)
    return int_type & mask


def toggle_bit(int_type, offset):
    mask = 1 << offset
    return int_type ^ mask


def flatten(x):
    if isinstance(x, collections.Iterable):
        return [a for i in x for a in flatten(i)]
    else:
        return [x]


class CIPDataType(ABC):
    """
    Abstract Base Class for CIP Data Types. The subclasses will be added to a data type dictionary and
    based on the data type code, the value method will provide an appropriate representation of the
    data to the client program when it reads from a CIP device and the the from_value method will format
    python types to the proper byte representation to send to the CIP device on write operations
    """

    @staticmethod
    @abstractmethod
    def data_type_code() -> bytes:
        """pass"""

    def __init__(self):
        self._data_type_code = self.data_type_code()
        self.addition_info_length = 0
        self.additional_info = b''
        self.data = bytearray()
        self.size = len(self.data)
        self.instance_id = None
        self.variable_name = ''

    def __repr__(self):
        return str(self.value())

    @property
    def alignment(self) -> int:
        """
        Data types have a memory alignment that describe how they are stored in memory which needs to be
        accounted for when reading or writing structures. Most basic data types align to their size, but
        strings align to a single byte, arrays to the size of their members (except Booleans), and unions
        and structures to the largest alignment of their members. Those types will override this method
        :return:
        """
        return self.size

    def bytes(self):
        byte_value = self.data_type_code() + \
                     self.addition_info_length.to_bytes(1, 'little') + \
                     self.additional_info + \
                     self.data
        return byte_value

    def from_bytes(self, bytes_cip_data_type):
        self._data_type_code = bytes_cip_data_type[0:1]
        self.addition_info_length = int.from_bytes(bytes_cip_data_type[1:2], 'little')
        self.additional_info = bytes_cip_data_type[2:2 + self.addition_info_length]
        self.data = bytes_cip_data_type[2 + self.addition_info_length:]

    @abstractmethod
    def value(self):
        """pass Abstract Docstring to pass coverage"""

    @abstractmethod
    def from_value(self, value):
        """pass Abstract Docstring to pass coverage"""


def update_data_type_dictionary(data_type_dictionary):
    """
    This function creates a datatype dictionary from all of the subclasses that inherit from the
    CIP datatype abstract base class

    :param data_type_dictionary:
    :return:
    """
    for sub_class in CIPDataType.__subclasses__():
        data_type_dictionary.update({sub_class.data_type_code(): sub_class})


def _get_class_data_type_code(data_type: bytes) -> CIPDataType:
    """
    This function will return a CIP datatype instance from a datatype code

    :param data_type:
    :return:
    """
    for sub_class in CIPDataType.__subclasses__():
        if sub_class.data_type_code() == data_type:
            return sub_class()


class CIPBoolean(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = struct.pack("<h", 0)
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xc1'  # (bit)

    def value(self):
        bool_value = int.from_bytes(self.data, 'little')
        if bool_value == 1:
            return True
        else:
            return False

    def from_value(self, value):
        if value:
            self.data = struct.pack("<h", 1)
        else:
            self.data = struct.pack("<h", 0)


class CIPShortInteger(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xc2'  # (1-byte signed binary)

    def value(self):
        return struct.unpack("<b", self.data)[0]

    def from_value(self, value):
        self.data = struct.pack("<b", value)


class CIPInteger(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xc3'  # (1-word signed binary)

    def value(self):
        return struct.unpack("<h", self.data)[0]

    def from_value(self, value):
        self.data = struct.pack("<h", value)


class CIPDoubleInteger(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00\x00\x00\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xc4'  # (2-word signed binary)

    def value(self):
        return struct.unpack("<l", self.data)[0]

    def from_value(self, value):
        self.data = struct.pack("<l", value)


class CIPLongInteger(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00\x00\x00\x00\x00\x00\x00\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xc5'  # (1-byte signed binary)

    def value(self):
        return struct.unpack("<q", self.data)[0]

    def from_value(self, value):
        self.data = struct.pack("<q", value)


class CIPUnsignedShortInteger(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xc6'  # (1-byte unsigned binary)

    def value(self):
        return struct.unpack("<B", self.data)[0]

    def from_value(self, value):
        self.data = struct.pack("<B", value)


class CIPUnsignedInteger(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xc7'  # (1-word unsigned binary)

    def value(self):
        return struct.unpack("<H", self.data)[0]

    def from_value(self, value):
        self.data = struct.pack("<H", value)


class CIPUnsignedDoubleInteger(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00\x00\x00\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xc8'  # (2-word unsigned binary)

    def value(self):
        return struct.unpack("<L", self.data)[0]

    def from_value(self, value):
        self.data = struct.pack("<L", value)


class CIPUnsignedLongInteger(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00\x00\x00\x00\x00\x00\x00\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xc9'  # (4-word unsigned binary)

    def value(self):
        return struct.unpack("<Q", self.data)[0]

    def from_value(self, value):
        self.data = struct.pack("<Q", value)


class CIPReal(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00\x00\x00\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xca'  # (2-word floating point)

    def value(self):
        return struct.unpack("<f", self.data)[0]

    def from_value(self, value):
        self.data = struct.pack("<f", value)


class CIPLongReal(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00\x00\x00\x00\x00\x00\x00\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xcb'  # (4-word floating point)

    def value(self):
        return struct.unpack("<d", self.data)[0]

    def from_value(self, value):
        self.data = struct.pack("<d", value)


class CIPString(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = str("").encode('utf-8')

    @staticmethod
    def data_type_code():
        return b'\xd0'  #

    @property
    def alignment(self) -> int:
        return 0

    def value(self):

        return str(self.data, 'utf-8').strip('\0')

    def from_value(self, value):
        byte_value = value.encode('utf-8')
        length_difference = self.size - len(byte_value)
        byte_value += length_difference * b'\x00'
        self.data = byte_value
        self.size = len(self.data)


class CIPByte(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xd1'

    def value(self):
        return self.data

    def from_value(self, value):
        self.data = value


class CIPWord(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xd2'  #

    def value(self):
        return self.data

    def from_value(self, value):
        self.data = value


class CIPDoubleWord(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00\x00\x00\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xd3'  #

    def value(self):
        return self.data

    def from_value(self, value):
        self.data = value


class CIPLongWord(CIPDataType):
    def __init__(self):
        super().__init__()
        self.data = b'\x00\x00\x00\x00\x00\x00\x00\x00'
        self.size = len(self.data)

    @staticmethod
    def data_type_code():
        return b'\xd4'  #

    def value(self):
        return self.data

    def from_value(self, value):
        self.data = value


class CIPTime(CIPDataType):
    def __init__(self):
        super().__init__()

    @staticmethod
    def data_type_code():
        return b'\xdb'  # (8-byte data)

    def value(self):
        pass

    def from_value(self, value):
        pass


class CIPAbbreviatedStructure(CIPDataType):
    def __init__(self):
        super().__init__()

    @staticmethod
    def data_type_code():
        return b'\xa0'  #

    def value(self):
        return self.data

    def from_value(self, value):
        self.data = value


class CIPStructure(CIPDataType):
    def __init__(self):
        super().__init__()
        self.variable_type_name = ''
        self.members = {}
        self._alignment = 0
        self.crc_code = b''

    @staticmethod
    def data_type_code():
        return b'\xa2'  #

    @property
    def alignment(self) -> int:
        return self._alignment

    @alignment.setter
    def alignment(self, new_alignment):
        self._alignment = new_alignment

    def __setitem__(self, key, value):
        current_type = self.members.get(key)
        current_type.from_value(value)
        self.members[key] = current_type
        self.from_value(self)

    def __getitem__(self, item):
        return self.members[item]

    def __repr__(self):
        return 'name: %s | type: %s | members: %s' % (self.variable_name, self.variable_type_name, self.members)

        if member.alignment > self.alignment:
            self.alignment = member.alignment
        alignment_offset = len(self.data) % member.alignment
        alignment_padding = 0
        if alignment_offset != 0:
            alignment_padding = member.alignment - alignment_offset
        self.data += b'\x00' * alignment_padding + member.data

    def add_member(self, member_name: str, member: CIPDataType):
        self.members[member_name] = member
        if member.alignment > self.alignment:
            self.alignment = member.alignment

    def value(self):
        offset = 0
        structure_data = self.data
        for member in self.members:
            member_value = self.members[member]
            if member_value.alignment != 0 and offset % member_value.alignment != 0:
                offset = offset + (member_value.alignment - offset % member_value.alignment)
            end_byte = offset + member_value.size
            member_value.data = structure_data[offset:end_byte]
            # Call value to handle nested structures
            member_value.value()
            offset = end_byte
        return self

    def from_value(self, value):
        # ToDo lookup data check
        offset = 0
        self.crc_code = value.crc_code
        mutable_data = bytearray(self.data)
        for member_key in value.members:
            member = value.members.get(member_key)
            if member.alignment != 0 and offset % member.alignment != 0:
                offset = offset + (member.alignment - offset % member.alignment)
            end_byte = offset + member.size
            mutable_data[offset:offset+member.size] = member.data
            offset = end_byte
        self.data = bytes(mutable_data)
        self.value()


class CIPArray(CIPDataType):
    # ToDo arrays of Boolean data types are a special case as they can pack 16 bits in their alignment
    def __init__(self):
        super().__init__()
        self.array_data_type = b''
        self.array_data_type_size = 0
        self.array_dimensions = 0
        self.number_of_elements = []
        self.start_array_elements = []
        self.size = 0
        self._local_cip_data_type_object = None
        self._list_representation = []
        self._alignment = 0
        self.data = b''
        # for i in range(self.array_dimensions):
        #     self._list_representation.append([self._local_cip_data_type_object.value()] * self.number_of_elements[i])

    @property
    def alignment(self) -> int:
        return self._alignment

    def _get_size(self) -> int:
        """
        Determine the amount of data required to send and receive data,
        handling the special case of Boolean array packing
        :return:
        """
        total_size = 1
        last_number = None
        for number in self.number_of_elements:
            total_size = total_size * number
            last_number = number
        add_bytes = 0
        if self.array_data_type == b'\xc1':
            if last_number % 16 != 0:
                add_bytes = 2
            # print("Total %s size in bool %s" % (self.bytes(), total_size))
            total_size = total_size // 8 + add_bytes
            # print("Total %s size in bool %s" % (self.bytes(), total_size))
            # self.array_data_type_size = required_words * 2
        else:
            total_size = total_size * self.array_data_type_size
        return total_size

    def from_items(self, array_data_type, array_data_size, array_dimensions, number_of_elements, start_array_elements):
        self.array_data_type = array_data_type
        self.array_data_type_size = array_data_size
        self.array_dimensions = array_dimensions
        self.number_of_elements = number_of_elements
        self.start_array_elements = start_array_elements
        self.size = self._get_size()
        # ToDo calling get_class_type_data_code seems weird
        self._local_cip_data_type_object = _get_class_data_type_code(self.array_data_type)
        self._alignment = self._local_cip_data_type_object.alignment

    def from_instance(self, cip_instance: CIPDataType, array_data_size,
                      array_dimensions, number_of_elements, start_array_elements):
        self.array_data_type = cip_instance.data_type_code()
        self.array_data_type_size = array_data_size
        self.array_dimensions = array_dimensions
        self.number_of_elements = number_of_elements
        self.start_array_elements = start_array_elements
        self.size = self._get_size()
        # ToDo calling get_class_type_data_code seems weird
        self._local_cip_data_type_object = cip_instance
        self._alignment = self._local_cip_data_type_object.alignment

    def _recursive_data_to_array(self, dimension: int = 0, position: int = 0):
        """
        Takes the dimension value of the array and converts the data bytes
        returned into lists nested to the dimension depth
        """
        local_list = []
        elements_per_index = 1
        for element in range(dimension + 1, self.array_dimensions):
            elements_per_index = elements_per_index * self.number_of_elements[element]
        if dimension == self.array_dimensions - 1:
            temp_array = []
            for element in range(self.number_of_elements[dimension]):
                if self.array_data_type == b'\xc1':
                    start_bit = (position + element)
                    if test_bit(int.from_bytes(self.data, 'little'), start_bit):
                        integer_bool = 1
                    else:
                        integer_bool = 0
                    self._local_cip_data_type_object.data = integer_bool.to_bytes(2, 'little')
                else:
                    start_bytes = (position + element) * self.array_data_type_size
                    self._local_cip_data_type_object.data = \
                        self.data[start_bytes:start_bytes + self.array_data_type_size]
                temp_value = copy.deepcopy(self._local_cip_data_type_object)
                temp_array.append(temp_value.value())
            return temp_array
        else:
            for element in range(self.number_of_elements[dimension]):
                local_list.append(
                    self._recursive_data_to_array(dimension + 1, position + elements_per_index * element))
            return local_list

    def _recursive_array_to_data(self, dimension: int, list_data):
        """
        This method takes the dimension value of the array object and recursively converts the list into byte data
        """

        if dimension == 1:
            data = b''
            for element in list_data:
                self._local_cip_data_type_object.from_value(element)
                data += self._local_cip_data_type_object.data
            return data
        else:
            data = b''
            for sub_list in list_data:
                data += self._recursive_array_to_data(dimension - 1, sub_list)
            return data

    @staticmethod
    def data_type_code():
        return b'\xa3'  # (1-byte signed binary) signed char

    def value(self):
        self._list_representation = self._recursive_data_to_array()
        return self._list_representation

    def from_value(self, value):
        if self.array_data_type == b'\xc1':
            flat_list = flatten(self._list_representation)
            data_int = 0
            offset = 0
            for item in flat_list:
                if item:
                    data_int = set_bit(data_int, offset)
                else:
                    data_int = clear_bit(data_int, offset)
                offset = offset + 1
            self.data = data_int.to_bytes(self.size, 'little')
        else:
            self.data = self._recursive_array_to_data(self.array_dimensions, value)


# class CIPDataTypes:
#     """
#     CIP has a byte that defines the data represented in the message
#     """
#     CIP_BOOLEAN = b'\xc1'  # (bit)
#     CIP_SINT = b'\xc2'  # (1-byte signed binary) signed char
#     CIP_INT = b'\xc3'  # (1-word signed binary) short
#     CIP_DINT = b'\xc4'  # (2-word signed binary) int
#     CIP_LINT = b'\xc5'  # (4-word signed binary) long long
#     CIP_USINT = b'\xc6'  # (1-byte unsigned binary)
#     CIP_UINT = b'\xc7'  # (1-word unsigned binary)
#     CIP_UDINT = b'\xc8'  # (2-word unsigned binary)
#     CIP_ULINT = b'\xc9'  # (4-word unsigned binary)
#     CIP_REAL = b'\xca'  # (2-word floating point)
#     CIP_LREAL = b'\xcb'  # (4-word floating point)
#     CIP_STRING = b'\xd0'
#     CIP_BYTE = b'\xd1'  # (1-byte hexadecimal)
#     CIP_WORD = b'\xd2'  # (1-word hexadecimal)
#     CIP_DWORD = b'\xd3'  # (2-word hexadecimal)
#     CIP_TIME = b'\xdb'  # (8-byte data)
#     CIP_LWORD = b'\xd4'  # (4-word hexadecimal)
#     CIP_ABBREVIATED_STRUCT = b'\xa0'
#     CIP_STRUCT = b'\xa2'
#     CIP_ARRAY = b'\xa3'
#     OMRON_UINT_BCD = b'\x04'  # (1-word unsigned BCD)
#     OMRON_UDINT_BCD = b'\x05'  # (2-word unsigned BCD)
#     OMRON_ULINT_BCD = b'\x06'  # (4-word unsigned BCD)
#     OMRON_ENUM = b'\x07'
#     OMRON_DATE_NSEC = b'\x08'
#     OMRON_TIME_NSEC = b'\x09'
#     OMRON_DATE_AND_TIME_NSEC = b'\x0a'
#     OMRON_TIME_OF_DAY_NSEC = b'\x0b'
#     OMRON_UNION = b'\x0c'
#
#     def __init__(self):
#         self.data_type_code = b''
#         self.addition_info_length = 0
#         self.additional_info = b''
#         self.data = b''
#
#     def bytes(self):
#         byte_value = self.data_type_code + \
#                      self.addition_info_length.to_bytes(1, 'little') + \
#                      self.additional_info + \
#                      self.data
#         return byte_value
#
#     def from_bytes(self, bytes_cip_data_type):
#         self.data_type_code = bytes_cip_data_type[0:1]
#         self.addition_info_length = int.from_bytes(bytes_cip_data_type[1:2], 'little')
#         self.additional_info = bytes_cip_data_type[2:2 + self.addition_info_length]
#         self.data = bytes_cip_data_type[2 + self.addition_info_length:]
#
#     def value(self):
#         if self.data_type_code == CIPDataTypes.CIP_BOOLEAN:
#             bool_value = int.from_bytes(self.data, 'little')
#             if bool_value == 1:
#                 return True
#             else:
#                 return False
#         elif self.data_type_code == CIPDataTypes.CIP_SINT:
#             return struct.unpack("<b", self.data)[0]
#         elif self.data_type_code == CIPDataTypes.CIP_INT:
#             return struct.unpack("<h", self.data)[0]
#         elif self.data_type_code == CIPDataTypes.CIP_DINT:
#             return struct.unpack("<l", self.data)[0]
#         elif self.data_type_code == CIPDataTypes.CIP_LINT:
#             return struct.unpack("<q", self.data)[0]
#         elif self.data_type_code == CIPDataTypes.CIP_USINT:
#             return struct.unpack("<B", self.data)[0]
#         elif self.data_type_code == CIPDataTypes.CIP_UINT:
#             return struct.unpack("<H", self.data)[0]
#         elif self.data_type_code == CIPDataTypes.CIP_UDINT:
#             return struct.unpack("<L", self.data)[0]
#         elif self.data_type_code == CIPDataTypes.CIP_ULINT:
#             return struct.unpack("<Q", self.data)[0]
#         elif self.data_type_code == CIPDataTypes.CIP_REAL:
#             return struct.unpack("<f", self.data)[0]
#         elif self.data_type_code == CIPDataTypes.CIP_LREAL:
#             return struct.unpack("<d", self.data)[0]
#         elif self.data_type_code == CIPDataTypes.CIP_STRING:
#             pass
#         elif self.data_type_code == CIPDataTypes.CIP_WORD:
#             pass
#         elif self.data_type_code == CIPDataTypes.CIP_DWORD:
#             pass
#         elif self.data_type_code == CIPDataTypes.CIP_TIME:
#             pass
#         elif self.data_type_code == CIPDataTypes.CIP_LWORD:
#             pass
#         elif self.data_type_code == CIPDataTypes.CIP_ABBREVIATED_STRUCT:
#             pass
#         elif self.data_type_code == CIPDataTypes.CIP_STRUCT:
#             pass
#         elif self.data_type_code == CIPDataTypes.CIP_ARRAY:
#             pass
#         elif self.data_type_code == CIPDataTypes.OMRON_UINT_BCD:
#             pass
#         elif self.data_type_code == CIPDataTypes.OMRON_UDINT_BCD:
#             pass
#         elif self.data_type_code == CIPDataTypes.OMRON_ULINT_BCD:
#             pass
#         elif self.data_type_code == CIPDataTypes.OMRON_ENUM:
#             pass
#         elif self.data_type_code == CIPDataTypes.OMRON_DATE_NSEC:
#             pass
#         elif self.data_type_code == CIPDataTypes.OMRON_TIME_NSEC:
#             pass
#         elif self.data_type_code == CIPDataTypes.OMRON_DATE_AND_TIME_NSEC:
#             pass
#         elif self.data_type_code == CIPDataTypes.OMRON_TIME_OF_DAY_NSEC:
#             pass
#         elif self.data_type_code == CIPDataTypes.OMRON_UNION:
#             pass
#
#     def from_value(self, value, data_type_code: bytes, additional_info: bytes):
#         self.data_type_code = data_type_code
#         self.addition_info_length = len(additional_info)
#         self.additional_info = additional_info
#         if self.data_type_code == CIPDataTypes.CIP_BOOLEAN:
#             if value:
#                 self.data = struct.pack("<h", 1)
#             else:
#                 self.data = struct.pack("<h", 0)
#         elif self.data_type_code == CIPDataTypes.CIP_SINT:
#             self.data = struct.pack("<b", value)
#         elif self.data_type_code == CIPDataTypes.CIP_INT:
#             self.data = struct.pack("<h", value)
#         elif self.data_type_code == CIPDataTypes.CIP_DINT:
#             self.data = struct.pack("<l", value)
#         elif self.data_type_code == CIPDataTypes.CIP_LINT:
#             self.data = struct.pack("<q", value)
#         elif self.data_type_code == CIPDataTypes.CIP_USINT:
#             self.data = struct.pack("<B", value)
#         elif self.data_type_code == CIPDataTypes.CIP_UINT:
#             self.data = struct.pack("<H", value)
#         elif self.data_type_code == CIPDataTypes.CIP_UDINT:
#             self.data = struct.pack("<L", value)
#         elif self.data_type_code == CIPDataTypes.CIP_ULINT:
#             self.data = struct.pack("<Q", value)
#         elif self.data_type_code == CIPDataTypes.CIP_REAL:
#             self.data = struct.pack("<f", value)
#         elif self.data_type_code == CIPDataTypes.CIP_LREAL:
#             self.data = struct.pack("<d", value)
#         elif self.data_type_code == CIPDataTypes.CIP_STRING:
#             pass
#         elif self.data_type_code == CIPDataTypes.CIP_WORD:
#             pass
#         elif self.data_type_code == CIPDataTypes.CIP_DWORD:
#             pass
#         elif self.data_type_code == CIPDataTypes.CIP_TIME:
#             pass
#         elif self.data_type_code == CIPDataTypes.CIP_LWORD:
#             pass
#         elif self.data_type_code == CIPDataTypes.CIP_ABBREVIATED_STRUCT:
#             pass
#         elif self.data_type_code == CIPDataTypes.CIP_STRUCT:
#             pass
#         elif self.data_type_code == CIPDataTypes.CIP_ARRAY:
#             pass
#         elif self.data_type_code == CIPDataTypes.OMRON_UINT_BCD:
#             pass
#         elif self.data_type_code == CIPDataTypes.OMRON_UDINT_BCD:
#             pass
#         elif self.data_type_code == CIPDataTypes.OMRON_ULINT_BCD:
#             pass
#         elif self.data_type_code == CIPDataTypes.OMRON_ENUM:
#             pass
#         elif self.data_type_code == CIPDataTypes.OMRON_DATE_NSEC:
#             pass
#         elif self.data_type_code == CIPDataTypes.OMRON_TIME_NSEC:
#             pass
#         elif self.data_type_code == CIPDataTypes.OMRON_DATE_AND_TIME_NSEC:
#             pass
#         elif self.data_type_code == CIPDataTypes.OMRON_TIME_OF_DAY_NSEC:
#             pass
#         elif self.data_type_code == CIPDataTypes.OMRON_UNION:
#             pass



__author__ = 'Joseph Ryan'
__license__ = "GPLv2"
__maintainer__ = "Joseph Ryan"
__email__ = "jr@aphyt.com"

from abc import ABC, abstractmethod
import aphyt.cip as ap
import binascii


def cip_crc16(data: bytes, poly=0xa001) -> bytes:
    data = bytearray(data)
    crc = 0x0000
    for byte in data:
        crc = crc ^ byte
        for _ in range(0, 8):
            carry = crc & 1
            crc >>= 1
            if carry:
                crc = crc ^ poly
    return crc.to_bytes(2, 'little')


class CIPRequest:
    """
    Class for assembling CIP requests

    ``1756-pm020_-en-p.pdf 17 of 94``

    * Read Tag Service (0x4c)
    * Read Tag Fragmented Service (0x52)
    * Write Tag Service (0x4d)
    * Write Tag Fragmented Service (0x53)
    * Read Modify Write Tag Service (0x4e)
    """

    def __init__(self,
                 request_service: bytes,
                 request_path: bytes,
                 request_data: bytes = b''):
        """ Constructor method

        :param request_service: bytes
        :param request_path: bytes
        :param request_data: bytes
        """
        self.request_service = request_service
        self.request_data = request_data
        # Length is in Words, so the byte length is divided in half
        self.request_path_size = len(request_path) // 2
        self.request_path = request_path

    @property
    def bytes(self) -> bytes:
        return \
            self.request_service + self.request_path_size.to_bytes(1, 'little') + \
            self.request_path + self.request_data


class CIPReply:
    """
    Class for parsing CIP replies
    """

    def __init__(self, reply_bytes: bytes):
        """

        :param reply_bytes:
        """
        self.reply_service = reply_bytes[0:1]
        self.reserved = reply_bytes[1:2]
        self.general_status = reply_bytes[2:3]
        self.extended_status_size = reply_bytes[3:4]
        # Research replies that use this. It's usually zero, so I am guessing it is in words (like the request)
        extended_status_byte_offset = int.from_bytes(self.extended_status_size, 'little') * 2
        self.extended_status = reply_bytes[4:extended_status_byte_offset]
        self.reply_data = reply_bytes[4 + extended_status_byte_offset:]

    @property
    def bytes(self) -> bytes:
        """
        The bytes in the CIP reply
        :return:
        """
        return \
            self.reply_service + self.reserved + self.general_status + \
            self.extended_status_size + self.extended_status + self.reply_data


class CIPCommonFormat:
    """
    CIP common format is a format that is used to pack data to be used in CIP messages
    Look at W506 page 341 of 570 for definition
    """

    def __init__(self,
                 data_type: bytes = b'',
                 additional_info_length: int = 0,
                 additional_info: bytes = b'',
                 data: bytes = b''):
        self.data_type = data_type
        self.additional_info_length = additional_info_length
        self.additional_info = additional_info
        self.data = data

    def __repr__(self):
        return str(self.data_type + self.additional_info_length.to_bytes(1, 'little') + self.additional_info + self.data)

    def from_bytes(self, bytes_cip_common_format):
        self.data_type = bytes_cip_common_format[0:1]
        self.additional_info_length = int.from_bytes(bytes_cip_common_format[1:2], 'little')
        self.additional_info = bytes_cip_common_format[2:2 + self.additional_info_length]
        self.data = bytes_cip_common_format[2 + self.additional_info_length:]


class CIPDispatcher(ABC):
    """
    CIPDispatcher is an abstract base class that has the basic methods and data required
    to send and receive CIP messages from a peer that is equipped to communicate using
    this protocol

    The key method that subclasses need to implement is 'execute_cip_command' which will
    send a CIPRequest over the CIP network and receive a CIPReply

    ToDo monitor when the read, write, get, set services have all moved to CIP objects mixins and get rid of them
    """

    def __init__(self):
        self.variables = {}
        self.user_variables = {}
        self.system_variables = {}
        self.data_type_dictionary = {}

    @abstractmethod
    def execute_cip_command(self, request: CIPRequest) -> CIPReply:
        pass

    def read_tag_service(self, tag_request_path, number_of_elements=1) -> CIPReply:
        read_tag_request = \
            CIPRequest(ap.CIPService.READ_TAG_SERVICE, tag_request_path, number_of_elements.to_bytes(2, 'little'))
        return self.execute_cip_command(read_tag_request)

    def write_tag_service(self, tag_request_path, request_service_data: CIPCommonFormat, number_of_elements=1):
        data = request_service_data.data_type + \
               int(request_service_data.additional_info_length).to_bytes(1, 'little') + \
               request_service_data.additional_info + number_of_elements.to_bytes(2, 'little') + \
               request_service_data.data
        write_tag_request = CIPRequest(ap.CIPService.WRITE_TAG_SERVICE, tag_request_path, data)
        return self.execute_cip_command(write_tag_request)

    def read_tag_fragmented_service(self, tag_request_path, offset, number_of_elements):
        data = tag_request_path + number_of_elements.to_bytes(2, 'little') + offset.to_bytes(4, 'little')
        read_tag_fragmented_request = \
            CIPRequest(ap.CIPService.READ_TAG_FRAGMENTED_SERVICE, tag_request_path, data)
        return self.execute_cip_command(read_tag_fragmented_request)

    def write_tag_fragmented_service(self, tag_request_path, cip_datatype_code, data, offset, number_of_elements=1):
        data = \
            cip_datatype_code + b'\x00' + number_of_elements.to_bytes(2, 'little') + \
            offset.to_bytes(4, 'little') + data
        write_tag_fragmented_request = CIPRequest(ap.CIPService.WRITE_TAG_FRAGMENTED_SERVICE, tag_request_path, data)
        return self.execute_cip_command(write_tag_fragmented_request)

    def get_attribute_all_service(self, tag_request_path):
        get_attribute_all_request = CIPRequest(ap.CIPService.GET_ATTRIBUTE_ALL, tag_request_path)
        return self.execute_cip_command(get_attribute_all_request)

    def get_attribute_single_service(self, tag_request_path):
        get_attribute_single_request = CIPRequest(ap.CIPService.GET_ATTRIBUTE_SINGLE, tag_request_path)
        return self.execute_cip_command(get_attribute_single_request)

    def set_attribute_single_service(self, tag_request_path):
        set_attribute_single_request = CIPRequest(ap.CIPService.SET_ATTRIBUTE_SINGLE, tag_request_path)
        return self.execute_cip_command(set_attribute_single_request)


def address_request_path_segment(class_id: bytes = None, instance_id: bytes = None,
                                 attribute_id: bytes = None, element_id: bytes = None) -> bytes:
    """
    This function is to create a request path using the class, instance, attribute and element values
    of the data that the programmer will access

    :param class_id: class id with low byte first
    :param instance_id:
    :param attribute_id:
    :param element_id
    :return:
    """
    # Logical segment  1756-pm020 16 of 94
    request_path_bytes = b''
    if class_id is not None:
        # 8-bit id uses b'\x20 16-bit uses b'\x21'
        if len(class_id) == 1:
            request_path_bytes += b'\x20' + class_id
        elif len(class_id) == 2:
            request_path_bytes += b'\x21\x00' + class_id
    if instance_id is not None:
        # 8-bit id uses b'\x24' 16-bit uses b'\x25'
        if len(instance_id) == 1:
            request_path_bytes += b'\x24' + instance_id
        elif len(instance_id) == 2:
            request_path_bytes += b'\x25\x00' + instance_id
    if attribute_id is not None:
        # 8-bit id uses b'\x30' 16-bit uses b'\x31'
        if len(attribute_id) == 1:
            request_path_bytes += b'\x30' + attribute_id
        elif len(attribute_id) == 2:
            request_path_bytes += b'\x31\x00' + attribute_id
    if element_id is not None:
        # 8-bit id uses b'\x28' 16-bit uses b'\x29' 32-bit uses b'\x2a'
        if len(element_id) == 1:
            request_path_bytes += b'\x28' + element_id
        elif len(element_id) == 2:
            request_path_bytes += b'\x29\x00' + element_id
        elif len(element_id) == 4:
            request_path_bytes += b'\x2a\x00' + element_id
    return request_path_bytes


def variable_request_path_segment(variable_name: str) -> bytes:
    """
    This function is to create a request path using the variable name of the data that the programmer will access

    :param variable_name: The name of the variable for the request path
    :return: Request path as a bytes object
    """
    # Symbolic segment
    # 1756-pm020_-en-p.pdf 17 of 94
    request_path_bytes = b'\x91' + len(variable_name).to_bytes(1, 'little') + variable_name.encode('utf-8')
    if len(request_path_bytes) % 2 != 0:
        request_path_bytes = request_path_bytes + b'\x00'
    return request_path_bytes



__author__ = 'Joseph Ryan'
__license__ = "GPLv2"
__maintainer__ = "Joseph Ryan"
__email__ = "jr@aphyt.com"

from aphyt.cip.cip_datatypes import CIPDataType
import datetime


class OmronDateAndTime(CIPDataType):
    #     OMRON_DATE_AND_TIME_NSEC = b'\x0a'
    @staticmethod
    def data_type_code():
        return b'\x0a'  #

    def value(self):
        unix_timestamp = int.from_bytes(self.data, 'little') / 10 ** 9
        datetime.datetime.fromtimestamp(unix_timestamp)
        return datetime.datetime.fromtimestamp(unix_timestamp, datetime.timezone.utc)

    def from_value(self, value):
        self.data = value


class OmronTime(CIPDataType):
    #     OMRON_TIME_NSEC = b'\x09'
    @staticmethod
    def data_type_code():
        return b'\x09'  #

    def value(self):
        return self.data

    def from_value(self, value):
        self.data = value


class OmronTimeOfDay(CIPDataType):
    #     OMRON_TIME_OF_DAY_NSEC = b'\x0b'
    @staticmethod
    def data_type_code():
        return b'\x0b'  #

    def value(self):
        return self.data

    def from_value(self, value):
        self.data = value


class OmronUnion(CIPDataType):
    #     OMRON_UNION = b'\x0c'
    @staticmethod
    def data_type_code():
        return b'\x0c'  #

    def value(self):
        return self.data

    def from_value(self, value):
        self.data = value


class OmronDate(CIPDataType):
    #     OMRON_DATE_NSEC = b'\x08'
    @staticmethod
    def data_type_code():
        return b'\x08'  #

    def value(self):
        return self.data

    def from_value(self, value):
        self.data = value


class OmronEnum(CIPDataType):
    #     OMRON_ENUM = b'\x07'
    @staticmethod
    def data_type_code():
        return b'\x07'  #

    def value(self):
        return self.data

    def from_value(self, value):
        self.data = value


class OmronUnsignedIntegerBCD(CIPDataType):
    #     OMRON_UINT_BCD = b'\x04'  # (1-word unsigned BCD)
    @staticmethod
    def data_type_code():
        return b'\x04'  #

    def value(self):
        return self.data

    def from_value(self, value):
        self.data = value


class OmronUnsignedDoubleInteger(CIPDataType):
    #     OMRON_UDINT_BCD = b'\x05'  # (2-word unsigned BCD)
    @staticmethod
    def data_type_code():
        return b'\x05'  #

    def value(self):
        return self.data

    def from_value(self, value):
        self.data = value


class OmronUnsignedLongInteger(CIPDataType):
    #     OMRON_ULINT_BCD = b'\x06'  # (4-word unsigned BCD)
    @staticmethod
    def data_type_code():
        return b'\x06'  #

    def value(self):
        return self.data

    def from_value(self, value):
        self.data = value



__author__ = 'Joseph Ryan'
__license__ = "GPLv2"
__maintainer__ = "Joseph Ryan"
__email__ = "jr@aphyt.com"

from .cip import *
from .cip_datatypes import *
from .cip_attributes import *
from .cip_services import *
